/*
 *
 *   qrash: the second portable demo in the world
 *
 *   Copyright (C) 1997  Queue Members Group Art Division
 *   Coded by Mad Max / Queue Members Group (Mike Shirobokov)
 *   <mad_max@qmg.rising.ru>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "video.h"
#include "misc.h"
#include "music.h"
#include <xtl.h>

PAGE current_page = 0;

IDirect3D8 *d3d;
IDirect3DDevice8 *device;
IDirect3DTexture8 *color_texture, *bw_texture; 
IDirect3DSurface8 *color_surface, *bw_surface;
IDirect3DPalette8 *palette; 
IDirect3DVertexBuffer8 *bg_vb;

#define D3DFVF_BACKGROUNDVERTEX (D3DFVF_XYZRHW|D3DFVF_TEX2)
struct BACKGROUNDVERTEX
{
    D3DXVECTOR4 p;
    FLOAT       tu1, tv1;
    FLOAT       tu2, tv2;
};

HRESULT rc;
#define check(x) 	{ debug( #x ); rc = x; if( rc != D3D_OK ) { char msg[256]; sprintf( msg, #x " failed = %lu", rc&0xFFFF ); error( msg ); } else { debug( "done" ); } }

void vidDoInitVideo() {
	debug( "vidDoInitVideo" );
	d3d = Direct3DCreate8( D3D_SDK_VERSION );
}

void vidSetVideoMode( int num ) {
	debug( "vidSetVideoMode: begin" );
	vidBytesPerLine = vidSizeX = num >> 16;
    vidSizeY = num & 0xFFFF;

    D3DPRESENT_PARAMETERS d3dpp; 
    ZeroMemory( &d3dpp, sizeof(d3dpp) );
    d3dpp.BackBufferWidth        = vidSizeX;
    d3dpp.BackBufferHeight       = vidSizeY;
	d3dpp.BackBufferFormat       = D3DFMT_X8R8G8B8;
    d3dpp.BackBufferCount        = 1;
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D24S8;
    d3dpp.SwapEffect             = D3DSWAPEFFECT_DISCARD;
    d3dpp.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_ONE_OR_IMMEDIATE;
    
	check( d3d->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, 0,
							  D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dpp, &device ) );

	check( device->CreatePalette( D3DPALETTE_256, &palette ) );

	check( D3DXCreateTexture( device, vidSizeX, vidSizeY, 1, 0, D3DFMT_P8, D3DPOOL_MANAGED, &color_texture ) );
	check( color_texture->GetSurfaceLevel( 0, &color_surface ) );

	check( D3DXCreateTexture( device, vidSizeX, vidSizeY, 1, 0, D3DFMT_L8, D3DPOOL_MANAGED, &bw_texture ) );
	check( bw_texture->GetSurfaceLevel( 0, &bw_surface ) );

	D3DSURFACE_DESC desc;
	check( color_texture->GetLevelDesc( 0, &desc ) );
	float tx_u = float( vidSizeX-1 ) / float( desc.Width );
	float tx_v = float( vidSizeY-1 ) / float( desc.Height );

	check( device->CreateVertexBuffer( 4*sizeof(BACKGROUNDVERTEX), 
									   D3DUSAGE_WRITEONLY, D3DFVF_BACKGROUNDVERTEX, D3DPOOL_DEFAULT, &bg_vb ) );
    BACKGROUNDVERTEX* vBackground;
    check( bg_vb->Lock( 0, 0, (BYTE**)&vBackground, 0 ) );
    
	float scale = 0.96f;

	vBackground[0].p = D3DXVECTOR4( float(vidSizeX)*(1.0f-scale), float(vidSizeY)*(1.0f-scale), 1.0f, 1.0f );
    vBackground[0].tu1 = 0.0f; vBackground[0].tv1 = 0.0f;
    vBackground[0].tu2 = 0.0f; vBackground[0].tv2 = 0.0f;
    
	vBackground[1].p = D3DXVECTOR4( float(vidSizeX)*(1.0f-scale), float(vidSizeY)*scale, 1.0f, 1.0f );
    vBackground[1].tu1 = 0.0f; vBackground[1].tv1 = tx_v;
    vBackground[1].tu2 = 0.0f; vBackground[1].tv2 = tx_v;
    
	vBackground[2].p = D3DXVECTOR4( float(vidSizeX)*scale, float(vidSizeY)*(1.0f-scale), 1.0f, 1.0f );
    vBackground[2].tu1 = tx_u; vBackground[2].tv1 = 0.0f;
    vBackground[2].tu2 = tx_u; vBackground[2].tv2 = 0.0f;

    vBackground[3].p = D3DXVECTOR4( float(vidSizeX)*scale, float(vidSizeY)*scale, 1.0f, 1.0f );
    vBackground[3].tu1 = tx_u; vBackground[3].tv1 = tx_v;
    vBackground[3].tu2 = tx_u; vBackground[3].tv2 = tx_v;

    check( bg_vb->Unlock() );
    
	check( device->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE ) );
	check( device->SetRenderState( D3DRS_LIGHTING, FALSE ) );
    check( device->SetRenderState( D3DRS_ZENABLE,	FALSE ) );

	vidPageSize = vidBytesPerLine*vidSizeY;
	vidMaxPage = 1;
	current_page = vidAllocPage();
	vidInitialized = true;
	debug( "vidSetVideoMode: end" );
}

void vidDoCloseVideo() {
    d3d->Release();
}

int vidChooseVideoMode() {
  return (640<<16)+480;
}

void vidDoSetPalette( vidRGB* pal, int border )
{
	debug( "vidDoSetPalette: begin" );
	D3DCOLOR *colors;
	check( palette->Lock( &colors, 0 ) );
    for( int i=0; i<256; i++ ) {
		colors[i] = D3DCOLOR_XRGB( pal[i].r, pal[i].g, pal[i].b );
//		colors[i] = D3DCOLOR_XRGB( i, i, i );
    }
	check( palette->Unlock() );
	device->SetPalette( 0, palette );
	debug( "vidDoSetPalette: end" );
}

void vidDoShowPage( PAGE color, PAGE bw, uchar* dither_table )
{
	debug( "vidDoShowPage: begin" );
//  vidDitherPage( color, bw, current_page, dither_table );
//  vidShowPage((PAGE)current_page);

	RECT r = { 0, 0, vidSizeX-1, vidSizeY-1 };		
	check( D3DXLoadSurfaceFromMemory( color_surface, NULL, &r, color, D3DFMT_P8, vidBytesPerLine, NULL, &r, D3DX_FILTER_NONE, 0 ) );
	check( D3DXLoadSurfaceFromMemory( bw_surface, NULL, &r, bw, D3DFMT_L8, vidBytesPerLine, NULL, &r, D3DX_FILTER_NONE, 0 ) );

    check( device->Clear( 0, NULL, D3DCLEAR_TARGET, 0x00000000, 1.0f, 0L ) );

	check( device->BeginScene() );

    check( device->SetRenderState( D3DRS_ZENABLE,			FALSE ) );
	check( device->SetRenderState( D3DRS_ZFUNC,				D3DCMP_ALWAYS ) );
    check( device->SetRenderState( D3DRS_FILLMODE,			D3DFILL_SOLID ) ); 
//    check( device->SetRenderState( D3DRS_FILLMODE,			D3DFILL_WIREFRAME ) ); 
    check( device->SetRenderState( D3DRS_ALPHABLENDENABLE,	FALSE ) ); 
    check( device->SetRenderState( D3DRS_ALPHATESTENABLE,	FALSE ) ); 
	check( device->SetRenderState( D3DRS_LIGHTING,			FALSE ) );
    check( device->SetPixelShader( 0 ) );

    // draw the background
    check( device->SetTexture( 0, color_texture ) );
    check( device->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1 ) );
    check( device->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE ) );
    check( device->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_DISABLE ) );
    check( device->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR ) );
    check( device->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR ) );
    check( device->SetTextureStageState( 0, D3DTSS_ADDRESSU,  D3DTADDRESS_CLAMP ) );
    check( device->SetTextureStageState( 0, D3DTSS_ADDRESSV,  D3DTADDRESS_CLAMP ) );

    // Blend in the lightmap map. 

    check( device->SetTexture( 1, bw_texture ) );
    check( device->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_MODULATE4X ) );
    check( device->SetTextureStageState( 1, D3DTSS_COLORARG1, D3DTA_TEXTURE ) );
    check( device->SetTextureStageState( 1, D3DTSS_COLORARG2, D3DTA_CURRENT ) );

/*
    check( device->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE ) );
    check( device->SetTextureStageState( 1, D3DTSS_MINFILTER, D3DTEXF_LINEAR ) );
    check( device->SetTextureStageState( 1, D3DTSS_MAGFILTER, D3DTEXF_LINEAR ) );
    check( device->SetTextureStageState( 1, D3DTSS_ADDRESSU,  D3DTADDRESS_CLAMP ) );
    check( device->SetTextureStageState( 1, D3DTSS_ADDRESSV,  D3DTADDRESS_CLAMP ) );
*/
    check( device->SetStreamSource( 0, bg_vb, sizeof(BACKGROUNDVERTEX) ) );
    check( device->SetVertexShader( D3DFVF_BACKGROUNDVERTEX ) );
    check( device->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ) );

	check( device->EndScene() );

	check( device->Present( NULL, NULL, NULL, NULL ) );
	debug( "vidDoShowPage: end" );
}

void vidMessage( char* msg )
{
	FILE* f = fopen( "D:\\qrash.log", "a" );
	fputs( msg, f ); fputc( '\n', f );
	fclose( f );

}
